%===============================================================================
%NOME: sintatica.pl
%
%Projeto: Correo de Programas
%
%   Verso: V 1.00
%   Autora : Juliana Carpes Imperial - 0124811-7
%
%Descrio do mdulo
%   Objetivos:
%      - implementao das clsulas que fazem a anlise sinttica da linguagem 
%        ALP, utilizando regras de DCG (Definite Clause Grammars); e
%      - implementao de um predicado que imprime o primeiro erro de sintaxe
%        encontrado, se for o caso. 
%===============================================================================


%===============================================================================
%Nome: fazerAnaSintatica
%Descrio do predicado: predicado principal do mdulo, o qual realiza todas as
%                        chamadas necessrias para realizar a anlise sinttica.
%
%Varivel de entrada:
%   - Tokens: a lista de tokens do cdigo fonte do programa em ALP.
%
%Varivel de sada:
%   - Arvore: a rvore sinttica do programa em ALP.
%===============================================================================

% analisa os tokens lidos do cdigo-fonte e monta a rvore sinttica 
% correspondente se o programa estiver sintticamente correto (ser abortado se
% no o estiver)
fazerAnaSintatica( Tokens, Arvore )                   :-
   write( '\nInterpretando: anlise sinttica...\n' ) ,
   serComandos( Arvore, Tokens, Tokens, [ eop ] )     ,
   write( 'Anlise sinttica feita com sucesso.\n' )  ,
   write( '\narvore sinttica = ' )                   ,
   write( Arvore )                                    ,
   nl                                                 ,
   !                                                  .


%===============================================================================
%Nome: serComandos
%Descrio da clusula: reconhece a lista de comandos do programa em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - Comandos: a lista de comandos dados no cdigo fonte se no houve erro
%               sinttico. Se houve, ser chamada a rotina de tratamento de 
%               erro e o programa, abortado.
%
%Clusula na BNF de ALP:
%   - < cmds > ::= < cmd > | < cmd > ; < cmds >
%===============================================================================

serComandos( Comandos, Tokens )              -->                                     
   serComando( Comando, Tokens )             ,   % reconhece primeiro comando
   serRestoCmds( Comando, Comandos, Tokens ) .   % reconhece os demais comandos

% se tentou esta clusula, houve erro sinttico, portanto, o erro ser tratado  
serComandos( _Comandos, Tokens ) -->
   tratarErro( Tokens )          .


%===============================================================================
%Nome: serRestoCmds
%Descrio da clusula: reconhece o restante da lista de comandos do programa em 
%                       ALP.
%
%Variveis de entrada: 
%   - Tokens:  a lista de tokens do cdigo fonte; e
%   - Comando: o ltimo comando reconhecido.
%
%Varivel de sada:
%   - Comandos: a lista de comandos reconhecidos, incluindo o ltimo comando, se
%               no houve erro sinttico; se houve, o programa ser abortado.
%===============================================================================

% reconhece o restante da lista de comandos, reconhecendo, antes, o ";" que
% veio depois de um comando j reconhecido
serRestoCmds( Comando, [ Comando | Comandos ], Tokens ) -->
   [ ptEVirg ]                                          ,    
   serComandos( Comandos, Tokens )                      .     

% no h mais comandos a serem reconhecidos; todos j o foram 
serRestoCmds( Comando, [ Comando ], _Tokens ) -->
   [ ]                                        . 


%===============================================================================
%Nome: serComando
%Descrio da clusula: reconhece um comando do programa em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - Comando: uma rvore com a estrutura do comando se no houve erro
%              sinttico. Se houve, ser chamada a rotina de tratamento de erro
%              e programa, abortado.
%
%Clusula na BNF de ALP:
%   - < cmd > ::= < var > := < exp >
%	             | skip
%               | if < bexp > then < cmds > fi
%               | if < bexp > then < cmds > else < cmds > esle_fi
%               | while < bexp > do < cmds > od
%===============================================================================

% reconhece um comando de atribuio, reconhecendo primeiro a varivel e, 
% posteriormente, a expresso que  atribuda a primeira
serComando( atribui( Variavel, Expressao ), Tokens ) -->
   [ ide( Variavel ), atribui ]                      ,   
   serExpressao( Expressao, Tokens )                 .    

% reconhece um comando skip
serComando( skip, _Tokens ) -->
   [ skip ]                 .

% reconhece um comando if, reconhecendo, nesta ordem, o identificador if, uma 
% expresso booleana, o identificador then, os comandos que sero executados
% caso o teste do if seja bem sucedido e, por fim, o resto do comando if
% (l a parte do else se ela existir)
serComando( Comando, Tokens )                  -->
   [ if ]                                      ,   
   serBoolExpr( Teste, Tokens )                ,   
   [ then ]                                    ,  
   serComandos( Entao, Tokens )                ,
   serRestoIf( Teste, Entao, Comando, Tokens ) .

% reconhece um comando while, reconhecendo, nesta ordem, o identificador 
% while, uma expresso booleana, o identificador do, os comandos do corpo
% do while e o identificador od, indicando fim do while
serComando( while( Teste, Comandos ), Tokens ) -->
   [ while ]                                   ,
   serBoolExpr( Teste, Tokens )                ,
   [ do ]                                      ,
   serComandos( Comandos, Tokens )             ,
   [ od ]                                      . 

% se tentou esta clusula, houve erro sinttico, portanto, o erro ser tratado  
serComando( _Comando, Tokens ) -->
   tratarErro( Tokens )        .
            

%===============================================================================
%Nome: serRestoIf
%Descrio da clusula: reconhece o restante de um comando if.
%
%Variveis de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte;
%   - Teste:  a expresso booleana a ser avaliada pelo comando if; e
%   - Ento:  a lista de comandos executada se o teste do if for bem sucedido.
%
%Varivel de sada:
%   - if( ... ): a rvore com a estrutura do comando if.
%===============================================================================

% se for um if-then-else, reconhece, nesta ordem, o identificador else, os 
% comandos que sero executados caso o teste do if no seja bem sucedido e o 
% e o identificador esle_fi, indicando fim do if-then-else
serRestoIf( Teste, Entao, if( Teste, Entao, Senao ), Tokens ) -->
   [ else ]                                                   , 
   serComandos( Senao, Tokens )                               ,
   [ esle_fi ]                                                .

% se for um if-then, reconhece apenas o identificador fi, indicando fim do 
% if-then, j que todo o comando if j foi reconhecido
serRestoIf( Teste, Entao, if( Teste, Entao ), _Tokens ) -->
   [ fi ]                                               .


%===============================================================================
%Nome: serExpressao
%Descrio da clusula: reconhece uma expresso numrica ou booleana do programa 
%                       em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - Expresso: uma rvore com a estrutura da expresso se no houve erro 
%                sinttico. Se houve, ser chamada a rotina de tratamento de 
%                erro e o programa, abortado.
%
%Clusula na BNF de ALP:
%   - < exp > ::= < bexp > | < iexp >
%===============================================================================

serExpressao( Expressao, Tokens )   --> 
   serIntExpr( Expressao )          ;   % reconhece expresso numrica, ou
   serBoolExpr( Expressao, Tokens ) .   % reconhece expresso booleana

% se tentou esta clusula, houve erro sinttico, portanto, o erro ser tratado  
serExpressao( _Expressao, Tokens ) -->
   tratarErro( Tokens )            .              


%===============================================================================
%Nome: serIntExpr
%Descrio da clusula: reconhece uma expresso numrica do programa em ALP.
%
%Varivel de sada:
%   - Expresso: uma rvore com a estrutura da expresso numrica se no houve 
%                erro sinttico. Se houve, o programa ser abortado.
%
%Clusula na BNF de ALP:
%   - < iexp > ::= < iterm > | < iterm > < fraco > < iexp >
%===============================================================================

serIntExpr( Expressao )                --> 
   serIntTermo( Termo )                ,   % reconhece o primeiro termo da expr
   serRestoIntExpr( Termo, Expressao ) .   % reconhece o resto da expr


%===============================================================================
%Nome: serRestoIntExpr
%Descrio da clusula: reconhece o resto de uma expresso numrica do programa 
%                       em ALP.
%
%Varivel de entrada:
%   - Termo: o ltimo termo reconhecido da expresso sendo reconhecida.
%
%Varivel de sada:
%   - Expresso: uma rvore com a estrutura da expresso numrica se no houve 
%                erro sinttico. Se houve, o programa ser abortado.
%===============================================================================

% reconhece o operador da expresso (+ / -), o termo seguinte, monta um termo
% com o termo anterior, o lido desta vez e o operador e, por fim, reconhece o 
% restante da expresso
serRestoIntExpr( Termo, Expressao )                                    --> 
   serOperFraco( Operador )                                            , 
   serIntTermo( OutroTermo )                                           , 
   serRestoIntExpr( expInt( Operador, Termo, OutroTermo ), Expressao ) .

% toda a expresso j foi reconhecida; nada mais precisa ser reconhecido
serRestoIntExpr( Expressao, Expressao ) --> 
   [ ]                                  .


%===============================================================================
%Nome: serIntTermo
%Descrio da clusula: reconhece um termo numrico do programa em ALP.
%
%Varivel de sada:
%   - Termo: uma rvore com a estrutura do termo numrico se no houve erro 
%            sinttico. Se houve, o programa ser abortado.
%
%Clusula na BNF de ALP:
%   - < iterm > ::= < ielem > | < ielem > < forte > < iterm >
%===============================================================================

serIntTermo( Termo )                   --> 
   serIntElem( Elemento )              ,   % reconhece o primeiro elem do termo
   serRestoIntTermo( Elemento, Termo ) .   % reconhece o resto do termo


%===============================================================================
%Nome: serRestoIntTermo
%Descrio da clusula: reconhece o resto de um termo numrico do programa em 
%                       ALP.
%
%Varivel de entrada:
%   - Elemento: o ltimo elemento reconhecido do termo sendo reconhecido.
%
%Varivel de sada:
%   - Termo: uma rvore com a estrutura do termo numrico se no houve erro 
%            sinttico. Se houve, o programa ser abortado.
%===============================================================================

% reconhece o operador do termo (* / /), o elemento seguinte, monta um elemento
% com o elemento anterior, o lido desta vez e o operador e, por fim, reconhece o 
% restante do termo
serRestoIntTermo( Elemento, Termo )                                   --> 
   serOperForte( Operador )                                           , 
   serIntElem( OutroElem )                                            , 
   serRestoIntTermo( expInt( Operador, Elemento, OutroElem ), Termo ) .

% todo o termo j foi reconhecido; nada mais precisa ser reconhecido
serRestoIntTermo( Termo, Termo ) --> 
   [ ]                           .


%===============================================================================
%Nome: serIntElem
%Descrio da clusula: reconhece um elemento numrico do programa em ALP.
%
%Varivel de sada:
%   - uma rvore com a estrutura do elemento numrico se no houve erro 
%     sinttico. Se houve, o programa ser abortado.
%
%Clusula na BNF de ALP:
%   - < ielem > ::= < num > 
%                 | < var > 
%                 | - ( < iexp > )
%                 | ( < iexp > )
%===============================================================================

% o elemento numrico reconhecido  um nmero inteiro
serIntElem( num( Num ) ) --> 
   [ num( Num ) ]        .

% o elemento numrico reconhecido  uma varivel
serIntElem( ide( Identificador ) ) -->
   [ ide( Identificador ) ]        .

% o elemento numrico reconhecido  um elemento com um sinal de menos na frente
serIntElem( menos( Elemento ) ) --> 
   [ menos ]                    ,   % reconhece sinal de menos 
   serIntElem( Elemento )       .   % reconhece o resto do elemento

% o elemento numrico reconhecido  uma expresso entre parnteses
serIntElem( Expressao )    --> 
   [ parenEsq ]            ,   % reconhece parntese esquerdo
   serIntExpr( Expressao ) ,   % reconhece expresso numrica
   [ parenDir ]            .   % reconhece parntese direito


%===============================================================================
%Nome: serBoolExpr
%Descrio da clusula: reconhece uma expresso booleana do programa em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - Expresso: uma rvore com a estrutura da expresso booleana se no houve 
%                erro sinttico. Se houve, ser chamada a rotina de tratamento 
%                de erro e o programa, abortado.
%
%Clusula na BNF de ALP:
%   - < bexp > ::= < bterm > | < bterm > or < bexp >
%===============================================================================

serBoolExpr( Expressao, Tokens )                --> 
   serBoolTermo( Termo, Tokens )                ,   % reconhece o primeiro termo
   serRestoBoolExpr( Termo, Expressao, Tokens ) .   % reconhece o resto da expr

% se tentou esta clusula, houve erro sinttico, portanto, o erro ser tratado  
serBoolExpr( _Expressao, Tokens ) -->
   tratarErro( Tokens )           .


%===============================================================================
%Nome: serRestoBoolExpr
%Descrio da clusula: reconhece o resto de uma expresso booleana do programa 
%                       em ALP.
%
%Variveis de entrada:
%   - Tokens: a lista de tokens do cdigo fonte; e
%   - Termo:  o ltimo termo reconhecido da expresso sendo reconhecida.
%
%Varivel de sada:
%   - Expresso: uma rvore com a estrutura da expresso booleana se no houve 
%                erro sinttico. Se houve, o programa ser abortado.
%===============================================================================

% reconhece o operador da expresso (or), o termo seguinte, monta um termo com o 
% termo anterior, o lido desta vez e o operador e, por fim, reconhece o restante 
% da expresso
serRestoBoolExpr( Termo, Expressao, Tokens )                               --> 
   [ or ]                                                                  , 
   serBoolTermo( OutroTermo, Tokens )                                      , 
   serRestoBoolExpr( expBool( or, Termo, OutroTermo ), Expressao, Tokens ) .

% toda a expresso j foi reconhecida; nada mais precisa ser reconhecido
serRestoBoolExpr( Expressao, Expressao, _Tokens ) --> 
   [ ]                                            .

   
%===============================================================================
%Nome: serBoolTermo
%Descrio da clusula: reconhece um termo booleano do programa em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - Termo: uma rvore com a estrutura do termo booleano se no houve erro 
%            sinttico. Se houve, o programa ser abortado.
%
%Clusula na BNF de ALP:
%   - < bterm > ::= < belem > | < belem > and < bterm >
%===============================================================================

serBoolTermo( Termo, Tokens )                   --> 
   serBoolElem( Elemento, Tokens )              ,   % reconhece o primeiro elem
   serRestoBoolTermo( Elemento, Termo, Tokens ) .   % reconhece o resto do termo


%===============================================================================
%Nome: serRestoBoolTermo
%Descrio da clusula: reconhece o resto de um termo booleano do programa em 
%                       ALP.
%
%Variveis de entrada:
%   - Tokens:   a lista de tokens do cdigo fonte; e
%   - Elemento: o ltimo elemento reconhecido do termo sendo reconhecido.
%
%Varivel de sada:
%   - Termo: uma rvore com a estrutura do termo booleano se no houve erro 
%            sinttico. Se houve, o programa ser abortado.
%===============================================================================

% reconhece o operador do termo (and), o elemento seguinte, monta um elemento
% com o elemento anterior, o lido desta vez e o operador e, por fim, reconhece o 
% restante do termo
serRestoBoolTermo( Elemento, Termo, Tokens )                               --> 
   [ and ]                                                                 , 
   serBoolElem( OutroElem, Tokens )                                        , 
   serRestoBoolTermo( expBool( and, Elemento, OutroElem ), Termo, Tokens ) .

% todo o termo j foi reconhecido; nada mais precisa ser reconhecido
serRestoBoolTermo( Termo, Termo, _Token ) --> 
   [ ]                                    .


%===============================================================================
%Nome: serBoolElem
%Descrio da clusula: reconhece um elemento booleano do programa em ALP.
%
%Varivel de entrada:
%   - Tokens: a lista de tokens do cdigo fonte.
%
%Varivel de sada:
%   - uma rvore com a estrutura do elemento booleano se no houve erro
%     sinttico. Se houve, o programa ser abortado.
%
%Clusulas na BNF de ALP:
%   - < belem > ::= true 
%                 | false 
%                 | < var > 
%                 | not ( < bexp > )
%                 | ( < bexp > )
%                 | < comp >
%   - < comp > ::= < iexp > < rel > < iexp > 
%===============================================================================

% o elemento booleano reconhecido  a constante true
serBoolElem( true, _Tokens ) --> 
   [ true ]                  .

% o elemento booleano reconhecido  a constante false
serBoolElem( false, _Tokens ) -->
   [ false ]                  .

% o elemento booleano reconhecido  uma varivel
serBoolElem( ide( Identificador ), _Tokens ) -->
   [ ide( Identificador ) ]                  .

% o elemento booleano reconhecido  um elemento negado
serBoolElem( neg( Elemento ), Tokens ) --> 
   [ not ]                             ,   % reconhece operador not
   serBoolElem( Elemento, Tokens )     .   % reconhece o resto do elemento

% o elemento booleano reconhecido  uma expresso entre parnteses
serBoolElem( Expressao, Tokens )    --> 
   [ parenEsq ]                     ,   % reconhece parntese esquerdo
   serBoolExpr( Expressao, Tokens ) ,   % reconhece expresso booleana
   [ parenDir ]                     .   % reconhece parntese direito

% o elemento booleano  uma comparao entre expresses numricas, onde uma
% expresso numrica  reconhecida, seguida de um operador de relao de ordem
% e, por fim,  reconhecida outra expresso numrica
serBoolElem( expBool( Relacao, Expressao_1, Expressao_2 ), _Tokens ) -->
   serIntExpr( Expressao_1 )                                         ,
   serRelacao( Relacao )                                             ,
   serIntExpr( Expressao_2 )                                         .


%===============================================================================
%Nome: serOperFraco
%Descrio da clusula: reconhece um operador de adio ou subtrao.
%
%Varivel de sada:
%   - o operador reconhecido.
%
%Clusula na BNF de ALP:
%   - < fraco > ::= + | -
%===============================================================================

serOperFraco( soma ) -->
   [ soma ]          .    % reconhece o operador de adio

serOperFraco( menos ) -->
   [ menos ]          .   % reconhece o operador de subtrao


%===============================================================================
%Nome: serOperForte
%Descrio da clusula: reconhece um operador de multiplicao ou diviso.
%
%Varivel de sada:
%   - o operador reconhecido.
%
%Clusula na BNF de ALP:
%   - < forte > ::= * | /
%===============================================================================

serOperForte( mult ) -->
   [ mult ]          .     % reconhece o operador de multiplicao

serOperForte( divide ) -->
   [ divide ]          .   % reconhece o operador de diviso


%===============================================================================
%Nome: serRelacao
%Descrio da clusula: reconhece um operador de comparao entre expresses
%                       numricas.
%
%Varivel de sada:
%   - o operador de relao reconhecido.
%
%Clusula na BNF de ALP:
%   - < rel > ::= <= | < | = | >= | <>
%===============================================================================

serRelacao( igual ) -->
   [ igual ]         .    % reconhece o operador de igualdade

serRelacao( difer ) -->
   [ difer ]         .    % reconhece o operador de desigualdade

serRelacao( menor ) -->
   [ menor ]        .     % reconhece o operador de menor do que

serRelacao( maior ) -->
   [ maior ]        .     % reconhece o operador de maior do que

serRelacao( menorIg ) -->
   [ menorIg ]        .   % reconhece o operador de menor ou igual do que

serRelacao( maiorIg ) -->
   [ maiorIg ]        .   % reconhece o operador de maior ou igual do que


%===============================================================================
%Nome: tratarErro
%Descrio do predicado: trata um erro encontrado durante a anlise sinttica.
%
%Variveis de entrada:
%   - Tokens:    a lista de tokens do cdigo fonte; e
%   - Restantes: os tokens que ainda no foram reconhecidos pelo analisador
%                sinttico.
%
%Varivel de sada:
%   - Intil: no retorna nada, j que o programa  abortado durante a execuo
%             deste predicado. Esta a apenas por causa da transformao das
%             clusulas DCG em predicados PROLOG.
%===============================================================================

% acha os tokens reconhecidos, pois  a primeira parte da lista dos tokens do
% cdigo fonte, cuja ltima parte so os que no foram reconhecidos por causa do
% erro
tratarErro( Tokens, Restantes, _Inutil ) :-
   nl                                    ,   
   write( 'Tokens reconhecidos: ' )      ,
   append( Lidos, Restantes, Tokens )    ,   % acha os tokens reconhecidos
   write( Lidos )                        ,   % imprime os reconhecidos
   nl                                    ,   
   write( 'Houve erro aqui: ' )          ,   % aponta onde houve o erro
   write( Restantes )                    ,   % imprime os no reconhecidos
   nl                                    ,   
   abort                                 .   % aborta o programa